from abc import abstractmethod
from datetime import datetime
from typing import Callable, Tuple

from sqlalchemy import Table

from sandwich import SANDWICH_VERSION
from sandwich.dialects import DialectHandler
from sandwich.modeling.dataclasses import ValidationResult
from sandwich.modeling.metadata import modeling_metadata

from .schema_generator import SchemaGenerator


class BaseSchemaGenerator(SchemaGenerator):
    """ Generates `job` and `drop` procedures\n
    Does not make any tables

    """
    def __init__(self, dialect_handler: DialectHandler, validation_result: ValidationResult, entity_registration_date: datetime):
        self.dialect_handler = dialect_handler
        self._validation_result = validation_result
        self.header = modeling_metadata.HEADER_TEMPLATE.format(
            created_on=entity_registration_date,
            updated_on=datetime.now(),
            version=SANDWICH_VERSION,
            entity_name=self._validation_result.entity_name
        )
        self._on_make_procedures: list[Callable[[dict[str, Tuple[str, str, str]], dict[str, Table]], None]] = []

    @abstractmethod
    def make_tables(self) -> dict[str, Table]:
        pass

    def make_procedures(self, tables: dict[str, Table]) -> dict[str, Tuple[str, str, str]]:
        procedures: dict[str, Tuple[str, str, str]] = {}

        # extension point
        for on_make_proc in self._on_make_procedures:
            on_make_proc(procedures, tables)

        proc_names = [tup[1] for tup in procedures.values()]
        table_schemas = list(procedures.keys())

        # job procedure
        job_proc_code, job_proc_name, job_call_stmt = self.dialect_handler.make_job_proc(
            entity_name=self._validation_result.entity_name,
            proc_names=proc_names,
            header=self.header
        )
        procedures["job"] = (job_proc_code, job_proc_name, job_call_stmt)
        proc_names.append(job_proc_name)

        # drop procedure
        drop_proc_code, drop_proc_name, drop_call_stmt = self.dialect_handler.make_drop_proc(
            entity_name=self._validation_result.entity_name,
            table_schemas=table_schemas,
            procedures=proc_names,
            header=self.header
        )
        procedures["drop"] = (drop_proc_code, drop_proc_name, drop_call_stmt)

        return procedures



from abc import ABC, abstractmethod
from typing import Tuple

from sqlalchemy import Table


class SchemaGenerator(ABC):
    @abstractmethod
    def make_tables(self) -> dict[str, Table]:
        pass

    @abstractmethod
    def make_procedures(self, tables: dict[str, Table]) -> dict[str, Tuple[str, str, str]]:
        pass



from dataclasses import dataclass
from datetime import datetime
from typing import Any, Tuple

from sqlalchemy import Table

@dataclass(frozen=True)
class Dv2Entity:
    entity_name: str
    profile: str
    created_on: datetime

@dataclass(frozen=True)
class StgInfo:
    stg_name: str
    stg_schema: str
    hk_keys: dict[str, Any]
    bk_keys: dict[str, Any]
    sys_columns: dict[str, Any]
    bus_columns: dict[str, Any]
    degenerate_field: dict[str, Any]

@dataclass(frozen=True)
class Dv2SystemInfo:
    entities_list: list[Dv2Entity]
    sys_entities: Table

@dataclass(frozen=True)
class ValidationResult:
    stg_schema: str
    entity_name: str
    bk_keys: list[Tuple[str, Any]]
    hk_keys: list[Tuple[str, Any]]
    business_column_types: dict[str, Any]
    system_column_types: dict[str, Any]
    profile: str
    degenerate_field: dict[str, Any]


from .factory import StrategyFactory
from .generators.schema_generator import SchemaGenerator
from .generators.base_schema_generator import BaseSchemaGenerator
from .generators.link_sat0_schema_generator import LinkSat0SchemaGenerator
from .generators.scd2_dim_schema_generator import Scd2DimSchemaGenerator
from .generators.hub_schema_generator import HubSchemaGenerator
from .validators.abc_validator import Validator
from .validators.base_validator import BaseValidator
from .validators.link_sat0_validator import LinkSat0Validator
from .validators.scd2_dim_validator import Scd2DimValidator
from .validators.hub_based_validator import HubBasedValidator


__all__ = [
    "Validator",
    "SchemaGenerator",
    "BaseValidator",
    "BaseSchemaGenerator",
    "HubBasedValidator",
    "HubSchemaGenerator",
    "StrategyFactory",
    "Scd2DimValidator",
    "Scd2DimSchemaGenerator",
    "LinkSat0Validator",
    "LinkSat0SchemaGenerator",
]





from datetime import datetime

from sandwich.dialects import DialectHandler
from sandwich.modeling.dataclasses import ValidationResult

from .generators.schema_generator import SchemaGenerator
from .validators.abc_validator import Validator
from .validators.link_sat0_validator import LinkSat0Validator
from .generators.link_sat0_schema_generator import LinkSat0SchemaGenerator
from .validators.scd2_dim_validator import Scd2DimValidator
from .generators.scd2_dim_schema_generator import Scd2DimSchemaGenerator
from .validators.hub_based_validator import HubBasedValidator
from .generators.hub_schema_generator import HubSchemaGenerator
from .validators.multi_sat2_validator import MultiSat2Validator
from .generators.multi_sat2_schema_generator import MultiSat2SchemaGenerator

class StrategyFactory:
    _strategies = {
        "hub-sat2-dim2": (Scd2DimValidator, Scd2DimSchemaGenerator),
        "link-sat0": (LinkSat0Validator, LinkSat0SchemaGenerator),
        "hub": (HubBasedValidator, HubSchemaGenerator),
        "multi-sat2": (MultiSat2Validator, MultiSat2SchemaGenerator),
    }

    @classmethod
    def register_strategy(cls, template_name: str, validator_class, generator_class):
        cls._strategies[template_name] = (validator_class, generator_class)

    @classmethod
    def create_validator(cls, profile: str) -> Validator:
        if profile not in cls._strategies:
            available = ", ".join(cls._strategies.keys())
            raise ValueError(f"Unknown profile '{profile}'. Available templates: {available}")

        validator_class, _ = cls._strategies[profile]
        return validator_class(profile)

    @classmethod
    def create_generator(cls, dialect_handler: DialectHandler, validation_result: ValidationResult, registered_on: datetime) -> SchemaGenerator:
        profile = validation_result.profile
        if profile not in cls._strategies:
            available = ", ".join(cls._strategies.keys())
            raise ValueError(f"Unknown profile '{profile}'. Available templates: {available}")

        _, generator_class = cls._strategies[profile]
        return generator_class(dialect_handler, validation_result, registered_on)

    @classmethod
    def get_available_templates(cls) -> list[str]:
        return list(cls._strategies.keys())



"""PostgreSQL dialect handler for SQL code generation."""
from typing import Tuple

from sqlalchemy import dialects, Table

from .base import DialectHandler
from .utils import get_columns_list


class PostgresDialectHandler(DialectHandler):
    """Dialect handler for PostgreSQL."""

    def get_boolean_type(self):
        return dialects.postgresql.BOOLEAN

    def get_proc_name_format(self, schema: str, operation: str, entity_name: str) -> str:
        """Get PostgreSQL procedure naming format."""
        return f'"{schema}"."{operation}_{entity_name}"'

    def apply_proc_template(self, proc_name: str, sql_body: str, header: str) -> str:
        """Wrap SQL body in PostgreSQL procedure profile with error handling."""
        # language=sql
        proc_template_sql = f"""
{header}
CREATE OR REPLACE PROCEDURE {proc_name}(p_parent_executionID BIGINT DEFAULT NULL)
LANGUAGE plpgsql
AS $$
DECLARE
    v_executionID BIGINT;
    v_ErrorMessage TEXT;
    v_ErrorDetail TEXT;
    v_ErrorContext TEXT;
    v_ErrorID INT;
BEGIN
    -- Log execution start
    CALL core."LogExecution"(0::INT, NULL::BIGINT, v_executionID, p_parent_executionID);

    BEGIN
        {sql_body}

        -- Log execution end
        CALL core."LogExecution"(0::INT, v_executionID, v_executionID, NULL::BIGINT);
    EXCEPTION WHEN OTHERS THEN
        GET STACKED DIAGNOSTICS
            v_ErrorMessage = MESSAGE_TEXT,
            v_ErrorDetail = PG_EXCEPTION_DETAIL,
            v_ErrorContext = PG_EXCEPTION_CONTEXT;

        -- Log error
        INSERT INTO core."ErrorLog" (
            "UserName",
            "ErrorNumber",
            "ErrorState",
            "ErrorSeverity",
            "ErrorLine",
            "ErrorProcedure",
            "ErrorMessage",
            "ErrorDateTime"
        ) VALUES (
            current_user,
            NULL,
            NULL,
            NULL,
            NULL,
            {proc_name}::TEXT,
            v_ErrorMessage || COALESCE(' | ' || v_ErrorDetail, ''),
            CURRENT_TIMESTAMP
        ) RETURNING "ErrorID" INTO v_ErrorID;

        -- Update execution log with error
        UPDATE core."ExecutionLog"
        SET "errorID" = v_ErrorID,
            "end_timestamp" = CURRENT_TIMESTAMP
        WHERE "executionID" = v_executionID;

        -- Re-raise the error
        RAISE;
    END;
END;
$$;
"""
        return proc_template_sql

    def make_stg_materialization_proc(self, entity_name: str, header: str) -> Tuple[str, str, str]:
        proc_name = self.get_proc_name_format("elt", f"Populate_stg", entity_name)

        # language=sql
        proc_body = f"""
        DROP TABLE IF EXISTS "stg"."{entity_name}";
        CREATE TABLE "stg"."{entity_name}" AS
        SELECT * FROM "proxy"."{entity_name}";
"""
        proc_code = self.apply_proc_template(proc_name, proc_body, header)
        return proc_code, proc_name, f"CALL {proc_name}()"

    def make_hub_proc(self, hub_table: Table, bk_keys: list, header: str) -> Tuple[str, str, str]:
        proc_name = self.get_proc_name_format("elt", f"Populate_{hub_table.schema}", hub_table.name)
        where_fields_list_str = " AND ".join([f'hub."{bk[0]}" = stg."{bk[0]}"' for bk in bk_keys])

        # Convert columns list to use double quotes
        columns_list = ", ".join([f'"{col.name}"' for col in hub_table.columns.values()])
        columns_list_stg = ", ".join([f'stg."{col.name}"' for col in hub_table.columns.values()])

        # language=sql
        proc_body = f"""
        INSERT INTO "{hub_table.schema}"."{hub_table.name}"
        ({columns_list})
        SELECT DISTINCT {columns_list_stg}
        FROM "stg"."{hub_table.name}" AS stg
        WHERE NOT EXISTS (
            SELECT 1
            FROM "{hub_table.schema}"."{hub_table.name}" AS hub
            WHERE {where_fields_list_str}
        );
"""
        proc_code = self.apply_proc_template(proc_name, proc_body, header)
        return proc_code, proc_name, f"CALL {proc_name}()"

    def make_link_proc(self, link_table: Table, hk_keys: list, header: str) -> Tuple[str, str, str]:
        proc_name = self.get_proc_name_format("elt", f"Populate_{link_table.schema}", link_table.name)
        where_fields_list_str = "\n\t\t\tAND ".join([f'link."{hk[0]}" = stg."{hk[0]}"' for hk in hk_keys if hk[0] != f"hk_{link_table.name}"])

        # Convert columns list to use double quotes
        columns_list = ", ".join([f'"{col.name}"' for col in link_table.columns.values()])
        columns_list_stg = ", ".join([f'stg."{col.name}"' for col in link_table.columns.values()])

        # language=sql
        proc_body = f"""
        INSERT INTO "{link_table.schema}"."{link_table.name}"
        ({columns_list})
        SELECT DISTINCT {columns_list_stg}
        FROM "stg"."{link_table.name}" AS stg
        WHERE NOT EXISTS (
            SELECT 1
            FROM "{link_table.schema}"."{link_table.name}" AS link
            WHERE {where_fields_list_str}
        );
"""
        proc_code = self.apply_proc_template(proc_name, proc_body, header)
        return proc_code, proc_name, f"CALL {proc_name}()"

    def make_scd0_sat_proc(self, sat_table: Table, header: str) -> Tuple[str, str, str]:
        proc_name = self.get_proc_name_format("elt", f"Populate_{sat_table.schema}", sat_table.name)
        hk_name = f"hk_{sat_table.name}"

        # Convert columns list to use double quotes
        columns_list = ", ".join([f'"{col.name}"' for col in sat_table.columns.values()])
        columns_list_stg = ", ".join([f'stg."{col.name}"' for col in sat_table.columns.values()])

        # language=sql
        proc_body = f"""
        INSERT INTO "{sat_table.schema}"."{sat_table.name}"
        ({columns_list})
        SELECT {columns_list_stg}
        FROM "stg"."{sat_table.name}" stg
        WHERE NOT EXISTS (
            SELECT 1
            FROM "sat"."{sat_table.name}" sat
            WHERE stg."{hk_name}" = sat."{hk_name}"
        );
"""
        proc_code = self.apply_proc_template(proc_name, proc_body, header)
        return proc_code, proc_name, f"CALL {proc_name}()"

    def make_scd2_sat_proc(
        self,
        sat_table: Table,
        hk_name: str,
        hashdiff_col: str,
        is_available_col: str,
        loaddate_col: str,
        stg_schema: str,
        header: str
    ) -> Tuple[str, str, str]:
        proc_name = self.get_proc_name_format("elt", f"Populate_{sat_table.schema}", sat_table.name)
        columns_list = ", ".join([f'"{col.name}"' for col in sat_table.columns.values()])

        def smart_replace(column_name: str) -> str:
            if column_name == "LoadDate":
                result = 'CURRENT_TIMESTAMP AS "LoadDate"'
            elif column_name == "IsAvailable":
                result = 'FALSE AS "IsAvailable"'
            else:
                result = f'sat."{column_name}"'
            return result

        select_columns_list = ", ".join([smart_replace(col.name) for col in sat_table.columns.values()])
        columns_list_stg = ", ".join([f'stg."{col.name}"' for col in sat_table.columns.values()])

        if stg_schema == "proxy":
            stg_table_name = f'"stg"."{sat_table.name}"'
            materialization_stmt = ""
        else:
            stg_table_name = "materialized_temp"
            materialization_stmt = f"""
        CREATE TEMPORARY TABLE materialized_temp AS
        SELECT DISTINCT {columns_list}
        FROM "stg"."{sat_table.name}";
"""

        # language=sql
        proc_body = f"""{materialization_stmt}
        WITH ranked_history AS (
            SELECT {columns_list},
                ROW_NUMBER() OVER (PARTITION BY "{hk_name}" ORDER BY "{loaddate_col}" DESC) AS "DescRank"
            FROM "{sat_table.schema}"."{sat_table.name}"
        )
        INSERT INTO "{sat_table.schema}"."{sat_table.name}"
        ({columns_list})
        SELECT {columns_list_stg}
        FROM {stg_table_name} stg
        WHERE NOT EXISTS (
            SELECT 1
            FROM ranked_history sat
            WHERE sat."DescRank" = 1
                AND stg."{hk_name}" = sat."{hk_name}"
                AND stg."{hashdiff_col}" = sat."{hashdiff_col}"
                AND sat."{is_available_col}" = TRUE
        )

        UNION ALL

        SELECT {select_columns_list}
        FROM ranked_history sat
        WHERE NOT EXISTS (
            SELECT 1
            FROM {stg_table_name} stg
            WHERE stg."{hk_name}" = sat."{hk_name}"
        )
        AND sat."DescRank" = 1
        AND sat."{is_available_col}" = TRUE;

        {f'DROP TABLE IF EXISTS {stg_table_name};' if stg_schema != "proxy" else ''}
"""
        proc_code = self.apply_proc_template(proc_name, proc_body, header)
        return proc_code, proc_name, f"CALL {proc_name}()"

    def make_scd2_dim_proc(self, dim_table: Table, bk_keys: list, header: str) -> Tuple[str, str, str]:
        proc_name = self.get_proc_name_format("elt", f"Recalculate_{dim_table.schema}", dim_table.name)
        columns_list = ", ".join([f'"{col.name}"' for col in dim_table.columns.values()])
        pk_keys = lambda: ", ".join([f'sat."{bk[0]}"' for bk in bk_keys])

        def smart_replace(column_name: str) -> str:
            if column_name == "DateFrom":
                result = 'sat."LoadDate" AS "DateFrom"'
            elif column_name == "DateTo":
                result = f'LEAD(sat."LoadDate" - INTERVAL \'1 microsecond\', 1, \'9999-12-31 23:59:59.999999\'::TIMESTAMP) OVER (PARTITION BY {pk_keys()} ORDER BY sat."LoadDate") AS "DateTo"'
            elif column_name == "IsCurrent":
                result = f'CASE WHEN LEAD(sat."LoadDate") OVER (PARTITION BY {pk_keys()} ORDER BY sat."LoadDate") IS NULL THEN TRUE ELSE FALSE END AS "IsCurrent"'
            else:
                result = f'sat."{column_name}"'
            return result

        select_columns_list = ",\n\t\t".join([smart_replace(col.name) for col in dim_table.columns.values()])

        # language=sql
        proc_body = f"""
        TRUNCATE TABLE "{dim_table.schema}"."{dim_table.name}";

        INSERT INTO "{dim_table.schema}"."{dim_table.name}"
        ({columns_list})
        SELECT {select_columns_list}
        FROM "sat"."{dim_table.name}" sat;
"""
        proc_code = self.apply_proc_template(proc_name, proc_body, header)
        return proc_code, proc_name, f"CALL {proc_name}()"

    def make_job_proc(self, entity_name: str, proc_names: list[str], header: str) -> Tuple[str, str, str]:
        proc_name = f'"job"."Run_all_related_to_{entity_name}"'
        proc_body = "\n\t\t"
        for proc in proc_names:
            if proc is None:
                continue
            proc_body += f"CALL {proc}(v_executionID);\n\t\t"

        proc_code = self.apply_proc_template(proc_name, proc_body, header)
        return proc_code, proc_name, f"CALL {proc_name}()"

    def make_drop_proc(self, entity_name, table_schemas: list[str], procedures: list[str], header: str) -> Tuple[str, str, str]:
        proc_name = f'"meta"."Drop_all_related_to_{entity_name}"'
        proc_body = "\n\t\t"
        for proc in procedures:
            if proc is None:
                continue
            proc_body += f"DROP PROCEDURE IF EXISTS {proc};\n\t\t"
        proc_body += "\n\t\t"
        for schema in table_schemas:
            proc_body += f'DROP TABLE IF EXISTS "{schema}"."{entity_name}";\n\t\t'
        proc_body += "\n\t\t"
        proc_body += f"""UPDATE core."entities"
        SET "deleted" = CURRENT_TIMESTAMP,
            "is_deleted" = TRUE
        WHERE "entity_name" = '{entity_name}';
"""

        proc_code = self.apply_proc_template(proc_name, proc_body, header)
        return proc_code, proc_name, f"CALL {proc_name}()"



from typing import Any

from sqlalchemy import Table, Engine, Connection, MetaData

from sandwich.modeling.dataclasses import StgInfo
from sandwich.modeling.metadata import modeling_metadata


def get_stg_info(entity_name: str, schema: str, conn: Engine | Connection) -> StgInfo:
    stg = Table(entity_name, MetaData(), schema=schema, autoload_with=conn)

    hk_keys: dict[str, Any] = {}
    bk_keys: dict[str, Any] = {}
    sys_columns: dict[str, Any] = {}
    bus_columns: dict[str, Any] = {}
    degenerate_field: dict[str, Any] = {}

    for col in stg.columns.values():
        if col.name.startswith("hk_"): # hash key
            hk_keys[col.name] = col.type
        elif col.name.startswith("bk_"): # business key
            bk_keys[col.name] = col.type
        elif col.name.startswith("dg_"): # degenerate field (transactional links only)
            degenerate_field[col.name] = col.type
        elif col.name.startswith("sg_"): # surrogate key
            raise Exception(f"sg column '{col.name}' is not implemented yet")
        elif col.name in modeling_metadata.names:
            # type_name = metadata.column_types[col.name]
            # if not str(col.type).startswith(type_name):
            #     raise Exception(f"{col.name} column must be `{type_name}` type, but it is `{str(col.type)}`")
            sys_columns[col.name] = col.type
        else:
            bus_columns[col.name] = col.type

    return StgInfo(
        stg_name=stg.name,
        stg_schema=stg.schema,
        hk_keys=hk_keys,
        bk_keys=bk_keys,
        sys_columns=sys_columns,
        bus_columns=bus_columns,
        degenerate_field=degenerate_field,
    )

def infer_profile(stg_info: StgInfo):
    hk_count = len(stg_info.hk_keys)
    hk_key = list(stg_info.hk_keys.items())[0]
    if hk_count == 0:
        raise Exception("hk column is required for any profile validation")
    elif hk_count > 1:
        return "link-sat0"
    else:
        if hk_key[0] != f"hk_{stg_info.stg_name}":
            return "multi-sat2"
        elif modeling_metadata.hashdiff in stg_info.sys_columns.keys():
            return "hub-sat2-dim2"
        else:
            return "hub"


from sandwich.modeling.dataclasses import Dv2SystemInfo, StgInfo

from .base_validator import BaseValidator
from ...metadata import modeling_metadata


class MultiSat2Validator(BaseValidator):
    """Validation rules:\n
    1. `bk_` column(s) should exist\n
    2. only one `hk_` column is allowed\n
    3. `hk_<entity_name>` column cant have `<entity_name>` = own entity name
    4. `HashDiff` and `IsAvailable` columns are required for scd2
    5. own entity name should start with another entity name as a prefix
    6. `hk_<entity_name>` should have entity_name = another entity name
    """
    def __init__(self, profile: str):
        super().__init__(profile)
        self._on_validate_staging.append(self._validate_multi_sat_staging)

    @staticmethod
    def _validate_multi_sat_staging(stg_info: StgInfo, sys_info: Dv2SystemInfo) -> None:
        if len(stg_info.bk_keys) == 0:
            raise Exception("bk column(s) are required for `multi-sat2` validation")

        hk_count = len(stg_info.hk_keys)
        if hk_count == 0:
            raise Exception("hk column is required for `multi-sat2` validation")
        elif hk_count > 1:
            raise Exception(f"More than one hk column found in stg.{stg_info.stg_name}")

        hk_key = list(stg_info.hk_keys.items())[0]
        if hk_key[0] == f"hk_{stg_info.stg_name}":
            raise Exception(f"hk column has invalid name '{hk_key[0]}'")

        system_column_names = stg_info.sys_columns.keys()
        if modeling_metadata.hashdiff not in system_column_names:
            raise Exception(f"{modeling_metadata.hashdiff} column is required for `multi-sat2` validation")
        if modeling_metadata.is_available not in system_column_names:
            raise Exception(f"{modeling_metadata.is_available} column is required for `multi-sat2` validation")

        other_entity_name: str | None = None
        for entity in sys_info.entities_list:
            if stg_info.stg_name.startswith(entity.entity_name) and stg_info.stg_name != entity.entity_name:
                other_entity_name = entity.entity_name
        if not other_entity_name:
            raise Exception("`multi-sat2` entity name should have other entity name as a suffix")

        if hk_key[0] != f"hk_{other_entity_name}":
            raise Exception(f"hk column has invalid name '{hk_key[0]}', correct name is `{other_entity_name}`")



from sandwich.modeling.dataclasses import Dv2SystemInfo, StgInfo

from .base_validator import BaseValidator


class HubBasedValidator(BaseValidator):
    def __init__(self, profile: str):
        super().__init__(profile)
        self._on_validate_staging.append(self._validate_hub_staging)

    @staticmethod
    def _validate_hub_staging(stg_info: StgInfo, sys_info: Dv2SystemInfo) -> None:
        # -----------------
        # hk
        # -----------------
        # only one hash key is allowed for a hub table
        # and its name should match `hk_[entity_name]` pattern
        hk_count = len(stg_info.hk_keys)
        if hk_count == 0:
            raise Exception("hk column is required for `hub-scd2dim` validation")
        elif hk_count > 1:
            raise Exception(f"More than one hk column found in stg.{stg_info.stg_name}")
        # hk_key = (key_name, key_type)
        hk_key = list(stg_info.hk_keys.items())[0]
        if hk_key[0] != f"hk_{stg_info.stg_name}":
            raise Exception(f"hk column has invalid name '{hk_key[0]}'")



from datetime import datetime
from typing import Tuple, Any

from sqlalchemy import Table, MetaData, Column, UniqueConstraint

from sandwich import SANDWICH_VERSION
from sandwich.dialects import DialectHandler
from sandwich.modeling.dataclasses import ValidationResult
from sandwich.modeling.metadata import modeling_metadata

from .schema_generator import SchemaGenerator


#todo: should be derived from `BaseSchemaGenerator`
class LinkSat0SchemaGenerator(SchemaGenerator):

    def __init__(self, dialect_handler: DialectHandler, validation_result: ValidationResult, entity_registration_date: datetime):
        self.dialect_handler = dialect_handler
        self._validation_result = validation_result
        self.header = modeling_metadata.HEADER_TEMPLATE.format(
            created_on=entity_registration_date,
            updated_on=datetime.now(),
            version=SANDWICH_VERSION,
            entity_name=self._validation_result.entity_name
        )

    def make_tables(self) -> dict[str, Table]:
        return {
            "link": self.make_link_table(),
            "sat": self.make_sat_table(),
            #"fact": self.make_fact_table(),
        }

    def make_link_table(self) -> Table:
        entity_name = self._validation_result.entity_name

        # Create link table
        link_table = Table(entity_name, MetaData(), schema="link")
        uks: list[str] = []

        # HKs (own and FKs)
        for hk_key in self._validation_result.hk_keys:
            if hk_key[0] == f"hk_{entity_name}":
                col = Column(hk_key[0], hk_key[1], primary_key=True)
            else:
                uks.append(hk_key[0])
                col = Column(hk_key[0], hk_key[1], nullable=False)
            link_table.append_column(col)

        for (dg_name, dg_type) in self._validation_result.degenerate_field.items():
            link_table.append_column(Column(dg_name, dg_type, nullable=False))
            uks.append(dg_name)

        link_table.append_constraint(UniqueConstraint(*uks))

        # LoadDate
        load_date = modeling_metadata.loaddate
        load_date_type = self._validation_result.system_column_types[load_date]
        load_date_col = Column(load_date, load_date_type, nullable=False)
        link_table.append_column(load_date_col)

        # RecordSource
        record_source = modeling_metadata.recordsource
        record_source_type = self._validation_result.system_column_types[record_source]
        record_source_col = Column(record_source, record_source_type, nullable=False)
        link_table.append_column(record_source_col)

        return link_table

    def make_sat_table(self) -> Table:
        entity_name = self._validation_result.entity_name

        # Create sat table
        sat_table = Table(entity_name, MetaData(), schema="sat")

        # own HK
        for hk_key in self._validation_result.hk_keys:
            if hk_key[0] == f"hk_{entity_name}":
                col = Column(hk_key[0], hk_key[1], primary_key=True)
                sat_table.append_column(col)

        # LoadDate
        load_date = modeling_metadata.loaddate
        load_date_type = self._validation_result.system_column_types[load_date]
        load_date_col = Column(load_date, load_date_type, nullable=False)
        sat_table.append_column(load_date_col)

        # RecordSource
        record_source = modeling_metadata.recordsource
        record_source_type = self._validation_result.system_column_types[record_source]
        record_source_col = Column(record_source, record_source_type, nullable=False)
        sat_table.append_column(record_source_col)

        # for transactional links
        for (dg_name, dg_type) in self._validation_result.degenerate_field.items():
            sat_table.append_column(Column(dg_name, dg_type, nullable=False))

        for (name_, type_) in self._validation_result.business_column_types.items():
            col = Column(name_, type_, nullable=True)
            sat_table.append_column(col)

        return sat_table

    def make_fact_table(self) -> Table:
        entity_name = self._validation_result.entity_name

        # Create a fact table
        fact_table = Table(entity_name, MetaData(), schema="fact")

        # not own HKs only
        for hk_key in self._validation_result.hk_keys:
            if hk_key[0] != f"hk_{entity_name}":
                col = Column(hk_key[0], hk_key[1], primary_key=True)
                fact_table.append_column(col)

        for (name_, type_) in self._validation_result.business_column_types.items():
            col = Column(name_, type_, nullable=True)
            fact_table.append_column(col)

        return fact_table

    def make_procedures(self, tables: dict[str, Table]) -> dict[str, Tuple[str, str, str]]:
        procedures = {}

        stg_proc_name = None
        if self._validation_result.stg_schema == "proxy":
            stg_proc_code, stg_proc_name, stg_call_stmt = self.dialect_handler.make_stg_materialization_proc(
                entity_name=self._validation_result.entity_name,
                header=self.header
            )
            procedures["stg"] = (stg_proc_code, stg_proc_name, stg_call_stmt)

        link_table = tables["link"]
        dg_keys: list[Tuple[str, Any]] = [(k, v) for k, v in self._validation_result.degenerate_field.items()]
        link_proc_code, link_proc_name, link_call_stmt = self.dialect_handler.make_link_proc(
            link_table=link_table,
            hk_keys=self._validation_result.hk_keys + dg_keys,
            header=self.header
        )
        procedures["link"] = (link_proc_code, link_proc_name, link_call_stmt)

        sat_table = tables["sat"]
        sat_proc_code, sat_proc_name, sat_call_stmt = self.dialect_handler.make_scd0_sat_proc(
            sat_table=sat_table,
            header=self.header
        )
        procedures["sat"] = (sat_proc_code, sat_proc_name, sat_call_stmt)

        # job procedure
        job_proc_names = [] # order-sensitive
        if self._validation_result.stg_schema == "proxy":
            job_proc_names.append(stg_proc_name)
        job_proc_names.extend([link_proc_name, sat_proc_name])
        job_proc_code, job_proc_name, job_call_stmt = self.dialect_handler.make_job_proc(
            entity_name=self._validation_result.entity_name,
            proc_names=job_proc_names,
            header=self.header
        )
        procedures["job"] = (job_proc_code, job_proc_name, job_call_stmt)

        # drop procedure
        drop_table_schemas = ["link", "sat"]
        if self._validation_result.stg_schema == "proxy":
            drop_table_schemas.append("stg")
        drop_proc_names = [link_proc_name, sat_proc_name, job_proc_name]
        if self._validation_result.stg_schema == "proxy":
            drop_proc_names.append(stg_proc_name)
        drop_proc_code, drop_proc_name, drop_call_stmt = self.dialect_handler.make_drop_proc(
            entity_name=self._validation_result.entity_name,
            table_schemas=drop_table_schemas,
            procedures=drop_proc_names,
            header=self.header
        )
        procedures["drop"] = (drop_proc_code, drop_proc_name, drop_call_stmt)

        return procedures


from datetime import datetime
from typing import Tuple, Any

from sqlalchemy import Table, MetaData, Column

from sandwich.dialects import DialectHandler
from sandwich.modeling.dataclasses import ValidationResult
from sandwich.modeling.metadata import modeling_metadata

from .base_schema_generator import BaseSchemaGenerator


class MultiSat2SchemaGenerator(BaseSchemaGenerator):

    def __init__(self, dialect_handler: DialectHandler, validation_result: ValidationResult, entity_registration_date: datetime):
        super().__init__(dialect_handler, validation_result, entity_registration_date)
        self._on_make_procedures.append(self._make_sat_proc)

    def make_tables(self) -> dict[str, Table]:
        return {
            "sat": self.make_sat_table(),
        }

    def make_sat_table(self) -> Table:
        entity_name = self._validation_result.entity_name

        # Create sat table
        sat_table = Table(entity_name, MetaData(), schema="sat")

        # BKs
        for bk_key in self._validation_result.bk_keys:
            col = Column(bk_key[0], bk_key[1], nullable=False)
            sat_table.append_column(col)

        # HK
        hk_key = self._validation_result.hk_keys[0]
        sat_table.append_column(Column(hk_key[0], hk_key[1], primary_key=True))

        # LoadDate
        load_date = modeling_metadata.loaddate
        load_date_type = self._validation_result.system_column_types[load_date]
        load_date_col = Column(load_date, load_date_type, primary_key=True)
        sat_table.append_column(load_date_col)

        # RecordSource
        record_source = modeling_metadata.recordsource
        record_source_type = self._validation_result.system_column_types[record_source]
        record_source_col = Column(record_source, record_source_type, nullable=False)
        sat_table.append_column(record_source_col)

        # HashDiff
        hashdiff_source = modeling_metadata.hashdiff
        hashdiff_type = self._validation_result.system_column_types[hashdiff_source]
        hashdiff_col = Column(hashdiff_source, hashdiff_type, nullable=False)
        sat_table.append_column(hashdiff_col)

        for (name_, type_) in self._validation_result.business_column_types.items():
            col = Column(name_, type_, nullable=True)
            sat_table.append_column(col)

        # IsAvailable
        is_available_source = modeling_metadata.is_available
        is_available_type = self._validation_result.system_column_types[is_available_source]
        is_available_col = Column(is_available_source, is_available_type, nullable=False)
        sat_table.append_column(is_available_col)

        return sat_table

    def _make_sat_proc(self, procedures: dict[str, Tuple[str, str, str]], tables: dict[str, Table]) -> None:
        sat_proc_code, sat_proc_name, sat_call_stmt = self.dialect_handler.make_scd2_sat_proc(
            sat_table=tables["sat"],
            hk_name=self._validation_result.hk_keys[0][0],
            hashdiff_col=modeling_metadata.hashdiff,
            is_available_col=modeling_metadata.is_available,
            loaddate_col=modeling_metadata.loaddate,
            stg_schema=self._validation_result.stg_schema,
            header=self.header
        )
        procedures["sat"] = (sat_proc_code, sat_proc_name, sat_call_stmt)


from sandwich.modeling.dataclasses import StgInfo, Dv2SystemInfo
from sandwich.modeling.metadata import modeling_metadata

from .hub_based_validator import HubBasedValidator


class Scd2DimValidator(HubBasedValidator):
    def __init__(self, profile: str):
        super().__init__(profile)
        self._on_validate_staging.append(self._validate_staging)

    @staticmethod
    def _validate_staging(stg_info: StgInfo, _: Dv2SystemInfo) -> None:

        # -----------------
        # BKs
        # -----------------
        # You don't need a hub or/and a dim tables for a non-business entity.
        # So you have to have at least one business key, and you can have more.
        # Naming convention is to just add a `bk_` prefix to the original key name
        # because we want to keep information of the original names
        if len(stg_info.bk_keys) == 0:
            raise Exception("bk column(s) are required for `hub-scd2dim` validation")


        system_column_names = stg_info.sys_columns.keys()

        if modeling_metadata.hashdiff not in system_column_names:
            raise Exception(f"{modeling_metadata.hashdiff} column is required for scd2dim validation")
        if modeling_metadata.is_available not in system_column_names:
            raise Exception(f"{modeling_metadata.is_available} column is required for scd2dim validation")



from abc import ABC, abstractmethod
from typing import Tuple

from sqlalchemy import Table, TextClause

class DialectHandler(ABC):
    @abstractmethod
    def get_boolean_type(self): ...

    @abstractmethod
    def get_proc_name_format(self, schema: str, operation: str, entity_name: str) -> str:
        pass

    @abstractmethod
    def apply_proc_template(self, proc_name: str, sql_body: str, header: str) -> str:
        pass

    @abstractmethod
    def make_stg_materialization_proc(
        self,
        entity_name: str,
        header: str
    ) -> Tuple[str, str, str]:
        pass

    @abstractmethod
    def make_hub_proc(
        self,
        hub_table: Table,
        bk_keys: list,
        header: str
    ) -> Tuple[str, str, str]:
        pass

    @abstractmethod
    def make_link_proc(
        self,
        link_table: Table,
        hk_keys: list,
        header: str
    ) -> Tuple[str, str, str]:
        pass

    @abstractmethod
    def make_scd2_sat_proc(
        self,
        sat_table: Table,
        hk_name: str,
        hashdiff_col: str,
        is_available_col: str,
        loaddate_col: str,
        stg_schema: str,
        header: str
    ) -> Tuple[str, str, str]:
        pass

    @abstractmethod
    def make_scd0_sat_proc(self, sat_table: Table, header: str) -> Tuple[str, str, str]:
        pass

    @abstractmethod
    def make_scd2_dim_proc(
        self,
        dim_table: Table,
        bk_keys: list,
        header: str
    ) -> Tuple[str, str, str]:
        pass

    @abstractmethod
    def make_job_proc(
        self,
        entity_name: str,
        proc_names: list[str],
        header: str
    ) -> Tuple[str, str, str]:
        pass

    @abstractmethod
    def make_drop_proc(self, entity_name, table_schemas: list[str], procedures: list[str], header: str) \
        -> Tuple[str, str, str]: ...


from sandwich.modeling.dataclasses import StgInfo, Dv2SystemInfo

from .base_validator import BaseValidator


class LinkSat0Validator(BaseValidator):
    def __init__(self, profile: str):
        super().__init__(profile)
        self._on_validate_staging.append(self._validate_staging)

    @staticmethod
    def _validate_staging(stg_info: StgInfo, sys_info: Dv2SystemInfo) -> None:
        # -----------------
        # hk
        # -----------------
        hk_count = len(stg_info.hk_keys)
        stg_full_name = f"stg.{stg_info.stg_name}"

        # so only exactly 2 or 3 columns allowed right now
        if hk_count < 2: # own and at least one foreign
            raise Exception(f"At least 2 hk columns expected in `{stg_full_name}` for the `link-sat0` profile")
        if hk_count > 3:
            raise Exception(f"{hk_count} hk columns in `{stg_full_name}` for the `link-sat0` profile?! Are you sure?")
        if len(stg_info.degenerate_field) == 0:
            raise Exception(f"At least one degenerate field is required for `{stg_full_name}` for the `link-sat0` profile")

        # hk_key = (key_name, key_type)
        hk_keys_copy = list(stg_info.hk_keys.keys()).copy()
        is_own_hk_found = False
        expected_own_hk_column_name = f"hk_{stg_info.stg_name}"

        for hk_name in stg_info.hk_keys.keys():
            if hk_name == expected_own_hk_column_name:
                hk_keys_copy.remove(hk_name)
                is_own_hk_found = True
            else:
                # check that `name` from `hk_[name]` is existing entity
                for en in sys_info.entities_list:
                    # for foreign keys
                    if hk_name == f"hk_{en.entity_name}":
                        hk_keys_copy.remove(hk_name)
                    # for a case when we reference the entity multiple times
                    # assuming foreign HKs are based on business columns from our link table
                    else:
                        for own_col in stg_info.bus_columns.keys():
                            if hk_name == f"hk_{en.entity_name}_{own_col}":
                                hk_keys_copy.remove(hk_name)

        if not is_own_hk_found:
            raise Exception(f"Column `{expected_own_hk_column_name}` has not been found in `{stg_full_name}`")

        if len(hk_keys_copy) > 0:
            raise Exception(f"There are no entities related to `{", ".join(hk_keys_copy)}` columns in the `{stg_full_name}`")

        # -----------------
        # BKs
        # -----------------
        # it should be a warning, not an error
        if len(stg_info.bk_keys) > 0:
            raise Exception("You dont need bk columns for the `link2fact` profile")





class Dv2ModelingMetadata:
    HEADER_TEMPLATE = """/* 
    =====================================================================
    AUTO-GENERATED CODE — DO NOT EDIT MANUALLY
    =====================================================================

    This stored procedure was automatically generated by:
      sandwich (https://pypi.org/project/sandwich/)

    Manual changes are discouraged.
    This file may be regenerated at any time, and all manual edits
    will be overwritten.

    Created on   : {created_on:%Y-%m-%d %H:%M:%S}
    Updated on   : {updated_on:%Y-%m-%d %H:%M:%S}
    Generator    : sandwich v{version}
    Entity name  : {entity_name}

    =====================================================================
*/
"""
    loaddate = "LoadDate"
    recordsource = "RecordSource"
    hashdiff = "HashDiff"
    is_available = "IsAvailable"
    names: list[str] = [loaddate, recordsource, hashdiff, is_available]

    # scenario format is "entity-context" or just "entity" if context is absent
    scenarios = ["original", "ecom_order_line-ecom_orders,ecom_products", "email_message-email_address",
                 "web_site_reg-email_address"]
    profiles = ["hub-sat2-dim2", "link-sat0", "hub", "multi-sat2"]
    supported_profile_upgrades = [("hub", "hub-sat2-dim2")]
    # _dialects_config: dict[str, dict[str, str]] = {
    #     "mssql": {
    #         loaddate: "DATETIME2",
    #         recordsource: "VARCHAR(200)",
    #         hashdiff: "CHAR(40)",
    #         is_available: "BIT",
    #     },
    #     "postgres": {
    #         loaddate: "TIMESTAMP",
    #         recordsource: "TEXT",
    #         hashdiff: "CHAR(40)",
    #         is_available: "BOOLEAN",
    #     },
    # }
    def __init__(self):
        #self.dialect = dialect

        #self.column_types = self._dialects_config[self.dialect]
        self.required_columns: list[str] = [self.loaddate, self.recordsource]

modeling_metadata = Dv2ModelingMetadata()


from datetime import datetime
from typing import Tuple

from sqlalchemy import Table, MetaData, Column, UniqueConstraint

from sandwich.dialects import DialectHandler
from sandwich.modeling.dataclasses import ValidationResult
from sandwich.modeling.metadata import modeling_metadata

from .base_schema_generator import BaseSchemaGenerator


class HubSchemaGenerator(BaseSchemaGenerator):
    def __init__(self, dialect_handler: DialectHandler, validation_result: ValidationResult, entity_registration_date: datetime):
        super().__init__(dialect_handler, validation_result, entity_registration_date)
        self._on_make_procedures.append(self._make_hub_proc)

    def make_tables(self) -> dict[str, Table]:
        return {
            "hub": self.make_hub_table(),
        }

    def make_hub_table(self) -> Table:
        entity_name = self._validation_result.entity_name

        hub_table = Table(entity_name, MetaData(), schema="hub")
        uks: list[str] = []

        # BKs
        for bk_key in self._validation_result.bk_keys:
            uks.append(bk_key[0])
            bk_col = Column(bk_key[0], bk_key[1], nullable=False)
            hub_table.append_column(bk_col)
        hub_table.append_constraint(UniqueConstraint(*uks))

        # HK
        hk_key = self._validation_result.hk_keys[0]
        hub_table.append_column(Column(hk_key[0], hk_key[1], primary_key=True))

        # LoadDate
        load_date = modeling_metadata.loaddate
        load_date_type = self._validation_result.system_column_types[load_date]
        load_date_col = Column(load_date, load_date_type, nullable=False)
        hub_table.append_column(load_date_col)

        # RecordSource
        record_source = modeling_metadata.recordsource
        record_source_type = self._validation_result.system_column_types[record_source]
        record_source_col = Column(record_source, record_source_type, nullable=False)
        hub_table.append_column(record_source_col)

        return hub_table

    def _make_hub_proc(self, procedures: dict[str, Tuple[str, str, str]], tables: dict[str, Table]) -> None:
        hub_proc_code, hub_proc_name, hub_call_stmt = self.dialect_handler.make_hub_proc(
            hub_table=tables["hub"],
            bk_keys=self._validation_result.bk_keys,
            header=self.header
        )
        procedures["hub"] = (hub_proc_code, hub_proc_name, hub_call_stmt)


from abc import ABC, abstractmethod

from sandwich.modeling.dataclasses import StgInfo, Dv2SystemInfo, ValidationResult


class Validator(ABC):
    @abstractmethod
    def validate_staging(self, stg_info: StgInfo, sys_info: Dv2SystemInfo, verbose: bool = False) -> ValidationResult:
        pass


# language=sql
create_extensions = """
create extension if not exists pgcrypto;
"""

# language=sql
create_entities_table = """
CREATE TABLE IF NOT EXISTS core.entities (
    entity_id BIGSERIAL PRIMARY KEY,
    entity_name VARCHAR(100) NOT NULL,
    profile VARCHAR(50) NOT NULL,
    created TIMESTAMP NOT NULL DEFAULT NOW(),
    updated TIMESTAMP NOT NULL DEFAULT NOW(),
    is_deleted BOOLEAN NOT NULL DEFAULT FALSE,
    deleted TIMESTAMP DEFAULT NULL,
    UNIQUE (entity_name)
);
"""

# language=sql
create_func_StringToHash= """
create or replace function core.string_to_hash1(str_value text)
returns char(40)
language plpgsql
as $$
declare
    result char(40);
begin
    if str_value is null
       or str_value in ('(unknown)', 'empty')
    then
        result := repeat('0', 40);
    else
        result :=
            upper(
                encode(
                    digest(upper(trim(str_value)), 'sha1'),
                    'hex'
                )
            );
    end if;

    return cast(result as char(40));
end;
$$;
"""

# language=sql
create_table_ExecutionLog = """
CREATE TABLE IF NOT EXISTS core."ExecutionLog" (
    "executionID" BIGSERIAL PRIMARY KEY,
    procid INT NOT NULL,
    begin_timestamp TIMESTAMP NOT NULL DEFAULT NOW(),
    end_timestamp TIMESTAMP DEFAULT NULL,
    "errorID" INT DEFAULT NULL,
    procname VARCHAR(200) DEFAULT NULL,
    "parent_executionID" BIGINT DEFAULT NULL
);
"""

# language=sql
create_table_ErrorLog = """
CREATE TABLE IF NOT EXISTS core."ErrorLog" (
    "ErrorID" SERIAL PRIMARY KEY,
    "UserName" VARCHAR(100) DEFAULT NULL,
    "ErrorNumber" INT DEFAULT NULL,
    "ErrorState" INT DEFAULT NULL,
    "ErrorSeverity" INT DEFAULT NULL,
    "ErrorLine" INT DEFAULT NULL,
    "ErrorProcedure" TEXT DEFAULT NULL,
    "ErrorMessage" TEXT DEFAULT NULL,
    "ErrorDateTime" TIMESTAMP DEFAULT NULL
);
"""

# language=sql
create_proc_LogExecution = """
CREATE OR REPLACE PROCEDURE core."LogExecution"(
    p_procid INT,
    "p_executionID_in" BIGINT,
    INOUT "p_executionID_out" BIGINT,
    "p_parent_executionID" BIGINT DEFAULT NULL
)
LANGUAGE plpgsql
AS $$
BEGIN
    IF "p_executionID_in" IS NOT NULL THEN
        UPDATE core."ExecutionLog"
        SET end_timestamp = NOW()
        WHERE "executionID" = "p_executionID_in";

        "p_executionID_out" := "p_executionID_in";
    ELSE
        INSERT INTO core."ExecutionLog" (procid, procname, "parent_executionID")
        VALUES (p_procid, NULL, "p_parent_executionID")
        RETURNING "executionID" INTO "p_executionID_out";
    END IF;
END;
$$;
"""

# language=sql
create_schemas = """
CREATE SCHEMA IF NOT EXISTS core;
CREATE SCHEMA IF NOT EXISTS stg;
CREATE SCHEMA IF NOT EXISTS hub;
CREATE SCHEMA IF NOT EXISTS sat;
CREATE SCHEMA IF NOT EXISTS link;
CREATE SCHEMA IF NOT EXISTS dim;
CREATE SCHEMA IF NOT EXISTS fact;
CREATE SCHEMA IF NOT EXISTS elt;
CREATE SCHEMA IF NOT EXISTS job;
CREATE SCHEMA IF NOT EXISTS meta;
CREATE SCHEMA IF NOT EXISTS proxy;
"""


from typing import Callable

from sandwich.modeling.dataclasses import StgInfo, Dv2SystemInfo, ValidationResult
from sandwich.modeling.metadata import modeling_metadata

from .abc_validator import Validator


class BaseValidator(Validator):
    """Validates required columns (LoadDate, RecordSource)\n
    Transforms `stg_info` input param into `ValidationResult` return value
    """
    def __init__(self, profile: str):
        self._on_validate_staging: list[Callable[[StgInfo, Dv2SystemInfo], None]] = []
        self.profile = profile

    def validate_staging(self, stg_info: StgInfo, sys_info: Dv2SystemInfo, verbose: bool = False) -> ValidationResult:
        """Raises: Exception"""
        if verbose:
            raise Exception("verbose is not implemented yet")

        system_column_names = stg_info.sys_columns.keys()

        # universal check - all dv2 raw objects should be auditable
        for required_col in modeling_metadata.required_columns:
            if required_col not in system_column_names:
                raise Exception(f"{required_col} column is required")

        # extension point
        for on_validate_staging in self._on_validate_staging:
            on_validate_staging(stg_info, sys_info)

        # todo: ValidationResult is not required whatsoever
        return ValidationResult(
            stg_schema=stg_info.stg_schema,
            entity_name=stg_info.stg_name,
            bk_keys=[(nm, tp) for nm, tp in stg_info.bk_keys.items()],
            hk_keys=[(nm, tp) for nm, tp in stg_info.hk_keys.items()],
            degenerate_field = stg_info.degenerate_field,
            business_column_types=stg_info.bus_columns,
            system_column_types=stg_info.sys_columns,
            profile=self.profile
        )
